#define _UNICODE            1
#define UNICODE             1
#define WIN32_LEAN_AND_MEAN 1
#include <windows.h>
#include <tchar.h>

#include "single.h"

#define AUDACIOUS_WND_CLASS_NAME    _T("AUDACIOUS_WND_CLASS_NAME")
#define AUDACIOUS_MAGIC_CODE        0x12345678

#define WM_SETUP_CALLBACK           (WM_APP + 0)

static DWORD  dwThread;
static HANDLE hThread;
static HANDLE hMutex;

static LRESULT APIENTRY
HandleCopyData(HWND hWnd, WPARAM wParam, LPARAM lParam)
{
#if _UNICODE
    LPSTR StrOut;
    int   Len;
#endif
    COPYDATASTRUCT *lpcds;
    single_instance_callback callback;

    lpcds = (COPYDATASTRUCT *)lParam;

    /* Small security system... */
    if (lpcds->dwData != AUDACIOUS_MAGIC_CODE)
        return 0;

    /* Recover callback to user-defined function */
    callback = (single_instance_callback)
               GetWindowLongPtr(hWnd, GWLP_USERDATA);

    /* If no callback, exit now */
    if (callback == NULL)
        return 0;

#if _UNICODE
    /* UNICODE version: convert to ASCII and post to callback */
    Len = WideCharToMultiByte (CP_ACP, 0, (LPWSTR)(lpcds->lpData), -1, 0, 0, NULL, NULL);

    StrOut = HeapAlloc(GetProcessHeap(), 0, Len * sizeof(char));
    if (!StrOut)
        return 0;

    WideCharToMultiByte (CP_ACP, 0, (LPWSTR)(lpcds->lpData), -1, StrOut, Len, 0, 0);

    callback(StrOut);

    HeapFree(GetProcessHeap(), 0, StrOut);
#else
    /* ASCII version: send transfer buffer directly. */
    callback((const char *)(lpcds->lpData));
#endif

    return 0;
}

static LRESULT APIENTRY
WindowProc(HWND hWnd, UINT message, WPARAM wParam, LPARAM lParam)
{
    switch (message) {
    case WM_COPYDATA:
        return HandleCopyData(hWnd, wParam, lParam);

    case WM_CLOSE:
        /* Close shadow window */
        DestroyWindow(hWnd);
        return 0L;

    case WM_DESTROY:
        /* Clean up and destroy handler */
        PostQuitMessage(0);
        return 0L;
    }

    return DefWindowProc(hWnd, message, wParam, lParam);
}

static HWND InitApp(HINSTANCE hInstance, int nShowCmd)
{
    TCHAR    szClass[] = AUDACIOUS_WND_CLASS_NAME;
    TCHAR    szTitle[] = _T("");
    HWND     hWnd;
    WNDCLASS wc;

    /* If the class doesn't exists, we must create it */
    if (!GetClassInfo(hInstance, szClass, &wc)) {
        wc.style         = 0;
        wc.lpfnWndProc   = WindowProc;
        wc.cbClsExtra    = 0;
        wc.cbWndExtra    = 0;
        wc.hInstance     = hInstance;
        wc.hIcon         = NULL;
        wc.hCursor       = NULL;
        wc.hbrBackground = NULL;
        wc.lpszMenuName  = NULL;
        wc.lpszClassName = szClass;
        RegisterClass(&wc);
    }

    /* Create shadow window */
    hWnd = CreateWindow(szClass,
                        szTitle,
                        WS_OVERLAPPEDWINDOW,
                        CW_USEDEFAULT,
                        CW_USEDEFAULT,
                        CW_USEDEFAULT,
                        CW_USEDEFAULT,
                        NULL,
                        NULL,
                        hInstance,
                        NULL);
    if (!hWnd)
        return FALSE;

    ShowWindow(hWnd, nShowCmd);
    UpdateWindow(hWnd);

    return hWnd;
}

static BOOL IsAppRunning(void)
{
    DWORD          dwRes;
    HWND           hWnd;
    COPYDATASTRUCT cds;
    LPTSTR         lpCommandLine;

    /* Create the global mutex */
    hMutex = CreateMutex(NULL, TRUE, AUDACIOUS_WND_CLASS_NAME);
    if (GetLastError() != ERROR_ALREADY_EXISTS)
        return FALSE;

    /* Wait that primary instance is initialized */
    dwRes = WaitForSingleObject(hMutex, 1000);

    /* Check if this secondary instance is valid */
    switch (dwRes) {
    case WAIT_OBJECT_0:
    case WAIT_ABANDONED:
        hWnd = FindWindow(AUDACIOUS_WND_CLASS_NAME, NULL);

        if (hWnd)
        {
            lpCommandLine = GetCommandLine();

            /* Numeric information (I use it as stupid "security") */
            cds.dwData = AUDACIOUS_MAGIC_CODE;
            /* Size of the transfer buffer */
            cds.cbData = (_tcslen(lpCommandLine) + 1) * sizeof(TCHAR);
            /* Command line to be sent to primary instance. */
            cds.lpData = lpCommandLine;

            SendMessage(hWnd, WM_COPYDATA, 0, (LPARAM)&cds);
        }

        break;
    }

    /* Close duplicated handle */
    CloseHandle(hMutex);

    return TRUE;
}

static int WINAPI WorkerThread(HANDLE hEvent)
{
    MSG       msg;
    HWND      hWnd;
    HINSTANCE hInstance;

    hInstance = GetModuleHandle(NULL);

    /* Forces creation of message queue */
    PeekMessage(&msg, NULL, WM_USER, WM_USER, PM_NOREMOVE);

    /* Create class and shadow window */
    hWnd = InitApp(hInstance, SW_HIDE);

    /* Initialization completed, signal to caller */
    SetEvent(hEvent);

    /* If no shadow window, exit now */
    if (hWnd == NULL)
    {
        return 1;
    }

    for (;;)
    {
        /* Wait for an event */
        if (GetMessage(&msg, NULL, 0, 0) <= 0)
            break;

        /* Check an exit command */
        if (msg.message == WM_CLOSE &&
            msg.hwnd    != hWnd)
        {
            PostMessage(hWnd, WM_CLOSE, 0, 0);
        } else
        /* Check if the user wants to set the callback */
        if (msg.message == WM_SETUP_CALLBACK)
            SetWindowLongPtr(hWnd, GWLP_USERDATA, msg.lParam);

        TranslateMessage(&msg);
        DispatchMessage(&msg);
    }

    return 0;
}

/* Starts the single instance mechanism.
 * It must be called as soon as the application starts.
 * Returns TRUE if the application must exit.
 */
int single_instance_start(void)
{
    HANDLE hEvent;

    if (IsAppRunning())
    {
        return 1;
    }

    /* Event for signaling initialization completed */
    hEvent = CreateEvent(NULL, TRUE, FALSE, NULL);
    if (hEvent == NULL)
        return 1;

    /* Start worker thread */
    hThread = CreateThread(
        NULL,
        0,
        (LPTHREAD_START_ROUTINE)
        WorkerThread,
        (LPVOID)hEvent,
        0,
        &dwThread);

    if (hThread == NULL)
        return 1;

    /* Wait for initialization completed */
    WaitForSingleObject(hEvent, INFINITE);

    /* Close that initialization event */
    CloseHandle(hEvent);

    /* Release ownership (it allows secondary instance) */
    ReleaseMutex(hMutex);

    return 0;
}

/* Registers a callback that it will be called
 * as soon as a secondary instance will be launched.
 * It will receive the command line sent on the secondary instance
 */
void single_instance_register(void *callback)
{
    PostThreadMessage(dwThread, WM_SETUP_CALLBACK, 0, (LPARAM)callback);
}

/* Stops the single instance mechanism */
void single_instance_stop(void)
{
    /* Send close command to shadow window and the worker thread */
    PostThreadMessage(dwThread, WM_CLOSE, 0, 0);

    /* Wait for action completed */
    WaitForSingleObject(hThread, INFINITE);

    /* Close thread handle */
    CloseHandle(hThread);

    /* Release the mutex */
    CloseHandle(hMutex);
}
